/*
 *  Copyright 2009 Heinrich Schuchardt.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *  under the License.
 */
package de.xypron.statistics;

/**
 * Exponentional distribution.
 */
public class ExponentialDistribution implements Distribution {

    /**
     * Random number generator.
     */
    private MersenneTwister generator = new MersenneTwister();
    /**
     * Mean.
     */
    private double mean;

    /**
     * Constructor.
     * @param my mean
     */
    public ExponentialDistribution(final double my) {
        setMean(my);
    }

    /**
     * Safety factor for alpha service level.
     * @param serviceLevel service level
     * @return safety factor
     * @throws Exception illegal value of service level
     */
    @Override
    public final double alphaSafetyFactor(final double serviceLevel)
            throws Exception {
        if (serviceLevel >= 1d) {
            throw new Exception("Service level too high");
        } else if (serviceLevel < 0d) {
            throw new Exception("Negative service level");
        } else if (serviceLevel == 0d) {
            return 0d;
        }
        return -Math.log(1d - serviceLevel) - 1d;
    }

    @Override
    public final double alphaServiceLevel(final double safetyFactor)
            throws Exception {
        if (safetyFactor < -1d) {
            return 0d;
        }
        return 1d - Math.exp(-1d - safetyFactor);
    }

    @Override
    public final double betaSafetyFactor(final double serviceLevel)
            throws Exception {
        if (serviceLevel >= 1) {
            throw new Exception("Service level too high");
        } else if (serviceLevel < 0) {
            throw new Exception("Negative service level");
        } else if (serviceLevel == 0.) {
            return 0.;
        }
        return -Math.log(1. - serviceLevel) - 1d;
    }

    @Override
    public final double betaServiceLevel(final double safetyFactor)
            throws Exception {
        if (safetyFactor < -1d) {
            return 0d;
        }
        return 1d - Math.exp(-1d - safetyFactor);
    }

    @Override
    public final double getMean() {
        return mean;
    }

    /**
     * Set mean value.
     * @param my mean
     */
    public final void setMean(final double my) {
        this.mean = my;
    }

    @Override
    public final double getStandardDeviation() {
        return mean;
    }

    @Override
    public final double randomValue() {
        return -Math.log(1. - generator.nextDouble()) * mean;
    }
}
