/*
 *  Copyright 2009 Heinrich Schuchardt.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *  under the License.
 */
package de.xypron.statistics;

/**
 * Gamma distribution.
 */
public class GammaDistribution implements Distribution {

    private static final double BIG = 4.503599627370496E15;
    private static final double BIGINV = 2.220446049250313E-16;
    private static final double MAXLOG = 709.782712893384;
    private static final double MACHEP = 1.1102230246251565E-16;
    /**
     * Random number generator.
     */
    private MersenneTwister generator = new MersenneTwister();
    /**
     * Mean.
     */
    private double my;
    /**
     * Standard deviation.
     */
    private double sigma;
    /**
     * Shape parameter.
     */
    private double kappa;
    /**
     * Scale parameter.
     */
    private double theta;

    /**
     * Constructor.
     * @param my mean
     * @param sigma standard deviation
     */
    public GammaDistribution(final double my, final double sigma) {
        this.my = my;
        this.sigma = sigma;

        init();
    }

    /**
     * Set mean demand.
     * @param my mean
     */
    public final void setMean(final double my) {
        this.my = my;
        init();
    }

    /**
     * Set standard deviation.
     * @param sigma standard deviation
     */
    public final void setStandardDeviation(final double sigma) {
        this.sigma = sigma;
        init();
    }

    private void init() {
        theta = sigma * sigma / my;
        kappa = my * my / sigma / sigma;
    }

    @Override
    public final double alphaSafetyFactor(final double serviceLevel)
            throws Exception {
        double r;
        r = (theta * gammaIncInv(serviceLevel, kappa) - my) / sigma;
        return r;
    }

    @Override
    public final double alphaServiceLevel(final double safetyFactor)
            throws Exception {
        double x = safetyFactor * sigma + my;
        return incompleteGamma(x / theta, kappa);
    }

    @Override
    public final double betaSafetyFactor(final double serviceLevel)
            throws Exception {
        double increment = 16;
        double ret = 0;

        do {
            if (betaServiceLevel(ret) > serviceLevel) {
                ret -= increment;
            } else {
                ret += increment;
            }
            increment = .5 * increment;
        } while (increment > 1e-7);
        return ret;
    }

    @Override
    public final double betaServiceLevel(final double safetyFactor)
            throws Exception {
        double x;
        x = safetyFactor * sigma + my;
        if (x < 0 ) {
            return 0;
        }
        return theta / my * incompleteGamma(x / theta, kappa + 1) * (kappa)
                + x / my * (1 - incompleteGamma(x / theta, kappa));
    }

    @Override
    public final double randomValue()
            throws Exception {
        double x;
        x = rgamma(kappa, theta);
        return x;
    }

    /**
     * Random variates from the standard exponential distribution.
     * @return random number
     */
    private double expRand() {
        return -Math.log(1. - generator.nextDouble());
    }

    /**
     * Inverse incomplete gamma function.
     * @param p propability
     * @param a form factor
     * @return inverse incomplete gamma function
     * @throws Exception illegal probability
     */
    private double gammaIncInv(final double p, final double a)
            throws Exception {
        double x, x1, x2;
        double dP, phi;
        double n;
        double step, step1;
        double nmax = 10000; // maximum number of iterations

        if (p >= 1.0) {
            throw new Exception("probability too high");
        } else if (p < 0.0) {
            throw new Exception("negative probability");
        } else if (p == 0.0) {
            return 0.0;
        }

        x1 = Math.exp((logGamma(a) + Math.log(p)) / a);
        x2 = -Math.log(1. - p) + logGamma(a);
        x = .5 * (x1 + x2);

        if (p < 0.05) {
            x = x1;
        }
        if (p > 0.95) {
            x = x2;
        }

        dP = 0.;
        for (n = 0; n < nmax; n++) {
            dP = p - incompleteGamma(x, a);

            if (Math.abs(dP) <= MACHEP * p) {
                break;
            }

            phi = gammaPdf(x, a, 1);

            step = dP / Math.max(2 * Math.abs(dP / x), phi);
            step1 = -((a - 1) / x - 1) * step * step / 4.0;

            if (Math.abs(step1) < Math.abs(step)) {
                step += step1;
            }

            if (x + step > 0) {
                x += step;
            } else {
                x /= 2.0;
            }
        }

        if (Math.abs(dP) > 1e-5) {
            throw new Exception("inverse failed to converge: a = " + a + ", p ="
                    + p);
        }
        return x;
    }

    /**
     * Probability density function for the Gamma distribution
     *
     * @param x value
     * @param a shape parameter (kappa)
     * @param b inverse scale parameter (1/theta)
     * @return probability
     */
    private double gammaPdf(final double x, final double a, final double b) {
        double r;
        if (x < 0) {
            return 0;
        }
        if (x == 0) {
            r = (a == 1) ? 1 / b : 0;
            return r;
        }
        if (a == 1) {
            r = Math.exp(-x / b) / b;
            return r;
        }
        r = Math.exp((a - 1) * Math.log(x / b) - x / b - logGamma(a)) / b;
        return r;
    }

    /** Get mean
     *
     * @return mean
     */
    @Override
    public final double getMean() {
        return my;
    }

    /** Get standard deviation
     *
     * @return standard deviation
     */
    @Override
    public final double getStandardDeviation() {
        return sigma;
    }

    /**
     * Returns the Incomplete Gamma function
     * @param x the integration end point.
     * @param a the parameter of the gamma distribution.
     * @return function value
     *
     * http://www.ssfnet.org/download/ssfnet_raceway-2.0.tar.gz
     *
     * Copyright (C) 1999 CERN - European Organization for Nuclear Research.
     * Permission to use, copy, modify, distribute and sell this software and
     * its documentation for any purpose is hereby granted without fee, provided
     * that the above copyright notice appear in all copies and that both that
     * copyright notice and this permission notice appear in supporting
     * documentation. CERN makes no representations about the suitability of
     * this software for any purpose. It is provided "as is" without expressed
     * or implied warranty.
     */
    private double incompleteGamma(final double x, final double a) {

        double ans, ax, c, r;

        if (x <= 0 || a <= 0) {
            return 0.0;
        }

        if (x > 1.0 && x > a) {
            return 1.0 - incompleteGammaComplement(x, a);
        }

        /* Compute  x**a * exp(-x) / gamma(a)  */
        ax = a * Math.log(x) - x - logGamma(a);
        if (ax < -MAXLOG) {
            return 0.0;
        }

        ax = Math.exp(ax);

        /* power series */
        r = a;
        c = 1.0;
        ans = 1.0;
        do {
            r += 1.0;
            c *= x / r;
            ans += c;
        } while (c / ans > MACHEP);

        return (ans * ax / a);

    }

    /**
     * Returns the Complemented Incomplete Gamma function.
     * @param a the parameter of the gamma distribution.
     * @param x the integration start point.
     * @return function value
     *
     * http://www.ssfnet.org/download/ssfnet_raceway-2.0.tar.gz
     *
     * Copyright (C) 1999 CERN - European Organization for Nuclear Research.
     * Permission to use, copy, modify, distribute and sell this software and
     * its documentation for any purpose is hereby granted without fee, provided
     * that the above copyright notice appear in all copies and that both that
     * copyright notice and this permission notice appear in supporting
     * documentation. CERN makes no representations about the suitability of
     * this software for any purpose. It is provided "as is" without expressed
     * or implied warranty.
     */
    private double incompleteGammaComplement(final double x, final double a) {
        double ans, ax, c, yc, r, t, y, z;
        double pk, pkm1, pkm2, qk, qkm1, qkm2;

        if (x <= 0 || a <= 0) {
            return 1.0;
        }

        if (x < 1.0 || x < a) {
            return 1.0 - incompleteGamma(x, a);
        }

        ax = a * Math.log(x) - x - logGamma(a);
        if (ax < -MAXLOG) {
            return 0.0;
        }

        ax = Math.exp(ax);

        /* continued fraction */
        y = 1.0 - a;
        z = x + y + 1.0;
        c = 0.0;
        pkm2 = 1.0;
        qkm2 = x;
        pkm1 = x + 1.0;
        qkm1 = z * x;
        ans = pkm1 / qkm1;

        do {
            c += 1.0;
            y += 1.0;
            z += 2.0;
            yc = y * c;
            pk = pkm1 * z - pkm2 * yc;
            qk = qkm1 * z - qkm2 * yc;
            if (qk != 0) {
                r = pk / qk;
                t = Math.abs((ans - r) / r);
                ans = r;
            } else {
                t = 1.0;
            }

            pkm2 = pkm1;
            pkm1 = pk;
            qkm2 = qkm1;
            qkm1 = qk;
            if (Math.abs(pk) > BIG) {
                pkm2 *= BIGINV;
                pkm1 *= BIGINV;
                qkm2 *= BIGINV;
                qkm1 *= BIGINV;
            }
        } while (t > MACHEP);

        return ans * ax;
    }

    /**
     * Logarithm of the gamma function.
     * Lanczos C., Journal S.I.A.M. Numerical Analysis Ser. B, Vol. 1 , 86, 1964
     * cited in
     * http://people.sc.fsu.edu/~burkardt/f77_src/asa245/asa245.f
     * absolute error unknown
     * 
     * @param x value
     * @return logarithm of the gamma function
     */
    private static double logGamma(final double x) {
        // Coefficients used in logGamma
        //   log(sqrt(2*pi)
        final double lnsqrt2pi = 0.9189385332046727E+00;
        //   Coefficients for approximation of logGamma
        final double c0 = 0.9999999999995183E+00;
        final double c1 = 676.5203681218835E+00;
        final double c2 = -1259.139216722289E+00;
        final double c3 = 771.3234287757674E+00;
        final double c4 = -176.6150291498386E+00;
        final double c5 = 12.50734324009056E+00;
        final double c6 = -0.1385710331296526E+00;
        final double c7 = 0.9934937113930748E-05;
        final double c8 = 0.1659470187408462E-06;
        double a;
        double b;

        a = (x - .5) * Math.log(x + 6.5) - (x + 6.5);
        b = c0
                + c1 / x
                + c2 / (x + 1)
                + c3 / (x + 2)
                + c4 / (x + 3)
                + c5 / (x + 4)
                + c6 / (x + 5)
                + c7 / (x + 6)
                + c8 / (x + 7);
        return a + Math.log(b) + lnsqrt2pi;
    }

    /**
     *  Random variates from the gamma distribution.
     *
     *  The algorithms are described in the following papers:
     *
     *    [1] Shape parameter a &gt;= 1.0  Algorithm GD in:
     *
     *    Ahrens, J.H. and Dieter, U. (1982).
     *    Generating gamma variates by a modified
     *    rejection technique.
     *    Comm. ACM, 25, 47-54.
     *
     *
     *    [2] Shape parameter 0 &lt; a &lt; 1. Algorithm GS in:
     *
     *    Ahrens, J.H. and Dieter, U. (1974).
     *    Computer methods for sampling from gamma, beta,
     *    poisson and binomial distributions.
     *    Computing, 12, 223-246.
     *
     *    Input: a = parameter (mean) of the standard gamma distribution.
     *    Output: a variate from the gamma(a)-distribution
     *
     *  This implementations is based on:
     *  RANLIB, Library of Fortran Routines for Random Number Generation
     *
     *  The copyright statement there is:
     *  <cite>
     *  Code that appeared  in an ACM publication is subject to their
     *  algorithms policy:
     *
     *  Submittal of  an  algorithm    for publication  in   one of   the  ACM
     *  Transactions implies that unrestricted use  of the algorithm within  a
     *  computer is permissible.   General permission  to copy and  distribute
     *  the algorithm without fee is granted provided that the copies  are not
     *  made  or   distributed for  direct   commercial  advantage.    The ACM
     *  copyright notice and the title of the publication and its date appear,
     *  and  notice is given that copying  is by permission of the Association
     *  for Computing Machinery.  To copy otherwise, or to republish, requires
     *  a fee and/or specific permission.
     *
     *  Krogh, F.  Algorithms  Policy.  ACM  Tran.   Math.  Softw.   13(1987),
     *  183-186.
     *
     *  We place the Ranlib code that we have written in the public domain.
     *  </cite>
     *
     * @param kappa shape parameter
     * @param theta scale parameter
     * @return random variate of the gamma function
     * @throws Exception exception
     */
    private double rgamma(final double kappa, final double theta)
            throws Exception {
        final double q1 = 4.166669E-2;
        final double q2 = 2.083148E-2;
        final double q3 = 8.01191E-3;
        final double q4 = 1.44121E-3;
        final double q5 = -7.388E-5;
        final double q6 = 2.4511E-4;
        final double q7 = 2.424E-4;
        final double a1 = 0.3333333;
        final double a2 = -0.250003;
        final double a3 = 0.2000062;
        final double a4 = -0.1662921;
        final double a5 = 0.1423657;
        final double a6 = -0.1367177;
        final double a7 = 0.1233795;
        final double e1 = 1.0;
        final double e2 = 0.4999897;
        final double e3 = 0.166829;
        final double e4 = 4.07753E-2;
        final double e5 = 1.0293E-2;
        final double sqrt32 = 5.656854;
        double sgamma, s2, s, d, t, x, u, r, q0, b, si, c, v, q, e, w, p;

        if (kappa < 0.0 || theta <= 0.0) {
            if (theta == 0.) {
                return 0.;
            }
            throw new Exception("failed");
        }

        if (kappa >= 1.0) {
            /*
            STEP  1:  RECALCULATIONS OF S2,S,D IF A HAS CHANGED
             */
            s2 = kappa - 0.5;
            s = Math.sqrt(s2);
            d = sqrt32 - 12.0 * s;
            /*
            STEP  2:  T=STANDARD NORMAL DEVIATE,
            X=(S,1/2)-NORMAL DEVIATE.
            IMMEDIATE ACCEPTANCE (I)
             */
            t = generator.nextGaussian();
            x = s + 0.5 * t;
            sgamma = x * x;
            if (t >= 0.0) {
                return theta * sgamma;
            }
            /*
            STEP  3:  U= 0,1 -UNIFORM SAMPLE. SQUEEZE ACCEPTANCE (S)
             */
            u = generator.nextDouble();
            if (d * u <= t * t * t) {
                return theta * sgamma;
            }
            /*
            STEP  4:  RECALCULATIONS OF Q0,B,SI,C IF NECESSARY
             */
            r = 1.0 / kappa;
            q0 = ((((((q7 * r + q6) * r + q5) * r + q4) * r + q3)
                    * r + q2) * r + q1) * r;
            /*
            APPROXIMATION DEPENDING ON SIZE OF PARAMETER A
            THE CONSTANTS IN THE EXPRESSIONS FOR B, SI AND
            C WERE ESTABLISHED BY NUMERICAL EXPERIMENTS
             */
            if (kappa > 13.022) {
                /*
                CASE 3:  A .GT. 13.022
                 */
                b = 1.77;
                si = 0.75;
                c = 0.1515 / s;
            } else if (kappa > 3.686) {
                /*
                CASE 2:  3.686 .LT. A .LE. 13.022
                 */
                b = 1.654 + 7.6E-3 * s2;
                si = 1.68 / s + 0.275;
                c = 6.2E-2 / s + 2.4E-2;
            } else {
                /*
                CASE 1:  A .LE. 3.686
                 */
                b = 0.463 + s + 0.178 * s2;
                si = 1.235;
                c = 0.195 / s - 7.9E-2 + 1.6E-1 * s;
            }
            /*
            STEP  5:  NO QUOTIENT TEST IF X NOT POSITIVE
             */
            if (x > 0) {
                /*
                STEP  6:  CALCULATION OF V AND QUOTIENT Q
                 */
                v = t / (s + s);
                if (Math.abs(v) > 0.25) {
                    q = q0 - s * t + 0.25 * t * t + (s2 + s2)
                            * Math.log(1.0 + v);
                } else {
                    q = q0 + 0.5 * t * t * ((((((a7 * v + a6) * v + a5) * v
                            + a4) * v + a3) * v + a2) * v + a1) * v;
                }
                /*
                STEP  7:  QUOTIENT ACCEPTANCE (Q)
                 */
                if (Math.log(1.0 - u) <= q) {
                    return theta * sgamma;
                }
            }
            /*
            STEP  8:  E=STANDARD EXPONENTIAL DEVIATE
            U= 0,1 -UNIFORM DEVIATE
            T=(B,SI)-DOUBLE EXPONENTIAL (LAPLACE) SAMPLE
             */
            for (;;) {
                e = expRand();
                u = generator.nextDouble();
                u += (u - 1.0);
                if (u < 0.0) {
                    t = b - si * e;
                } else {
                    t = b + si * e;
                }

                /*
                STEP  9:  REJECTION IF T .LT. TAU(1) = -.71874483771719
                 */
                if (t >= -0.7187449) {

                    /*
                    STEP 10:  CALCULATION OF V AND QUOTIENT Q
                     */
                    v = t / (s + s);
                    if (Math.abs(v) > 0.25) {
                        q = q0 - s * t + 0.25 * t * t + (s2 + s2)
                                * Math.log(1.0 + v);
                    } else {
                        q = q0 + 0.5 * t * t * ((((((a7 * v + a6) * v + a5)
                                * v + a4) * v + a3) * v + a2) * v + a1) * v;
                    }
                    /*
                    STEP 11:  HAT ACCEPTANCE (H) (IF Q NOT POSITIVE GO TO STEP 8)
                     */
                    if (q > 0.0) {
                        if (q > 0.5) {
                            w = Math.exp(q) - 1.0;
                        } else {
                            w = ((((e5 * q + e4) * q + e3) * q + e2) * q + e1)
                                    * q;
                        }

                        /*
                        IF T IS REJECTED, SAMPLE AGAIN AT STEP 8
                         */
                        if (c * Math.abs(u) <= w * Math.exp(e - 0.5 * t * t)) {
                            break;
                        }
                    }
                }
            }
            x = s + 0.5 * t;
            sgamma = x * x;
            return theta * sgamma;
        } else {
            // ALTERNATE METHOD FOR PARAMETERS A BELOW 1  (.3678794=EXP(-1.))
            b = 1.0 + 0.3678794 * kappa;
            for (;;) {
                p = b * generator.nextDouble();
                if (p < 1.0) {
                    sgamma = Math.exp(Math.log(p) / kappa);
                    if (expRand() >= sgamma) {
                        break;
                    }
                } else {
                    sgamma = -Math.log((b - p) / kappa);
                    if (expRand() >= (1.0 - kappa) * Math.log(sgamma)) {
                        break;
                    }
                }
            }
            return sgamma * theta;
        }
    }
}
