/*
 *  Copyright 2009, 2010, 2011 Heinrich Schuchardt.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *  under the License.
 */
package de.xypron.statistics;

/**
 * Normal distribution.
 */
public class NormalDistribution implements Distribution {

    private MersenneTwister generator = new MersenneTwister();
    private double my;    // mean
    private double sigma; // standard deviation

    /** 
     * Constructor for normal distribution.
     * @param my mean
     * @param sigma standard deviation
     */
    public NormalDistribution(double my, double sigma) {
        this.my = my;
        this.sigma = sigma;
    }

    @Override
    public final double alphaSafetyFactor(double serviceLevel) throws Exception {
        final double twoSqrt = 1.4142135623730950488016887242097d;
        return SpecialFunctions.erfInv(2d * serviceLevel - 1d) * twoSqrt;
    }

    @Override
    public final double alphaServiceLevel(double safetyFactor) throws Exception {
        final double halfSqrt = 0.70710678118654752440084436210485d;
        return .5d * (1d + SpecialFunctions.erf(safetyFactor * halfSqrt));
    }

    @Override
    public final double betaSafetyFactor(double serviceLevel) throws Exception {
        final double eps = 1e-12d;
        double sf = 0;
        double sl;
        double x1;
        double x2;

        x1 = -200; // reasonable minimum
        x2 = 200; // reasonable maximum

        sf = 0;
        while (x2 - x1 > eps) {
            sf = .5 * (x1 + x2);
            sl = betaServiceLevel(sf);
            if (sl > serviceLevel) {
                x2 = sf;
            } else {
                x1 = sf;
            }
        }
        return sf;
    }

    @Override
    public final double betaServiceLevel(double k) throws Exception {
        double x;
        x = my + sigma * k;
        return -sigma / my / Math.sqrt(2d * Math.PI) * Math.exp(-.5 * k * k)
                + .5 * (1d + SpecialFunctions.erf(k / Math.sqrt(2d)))
                + x / my / 2 * (1d - SpecialFunctions.erf(k / Math.sqrt(2d)));
    }

    @Override
    public final double getMean() {
        return my;
    }

    /**
     * Set mean.
     * @param my mean
     */
    public final void setMean(final double my) {
        this.my = my;
    }

    @Override
    public final double getStandardDeviation() {
        return sigma;
    }

    /**
     * Set standard deviation
     * @param sigma standard deviation 
     */
    public final void setStandardDeviation(final double sigma) {
        this.sigma = sigma;
    }

    @Override
    public final double randomValue() throws Exception {
        return my + sigma * generator.nextGaussian();
    }

}
