/*
 * Copyright 2012 Heinrich Schuchardt <xypron.glpk@gmx.de>.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.xypron.statistics;

/**
 * Rayleigh Distribution.
 * @author Heinrich Schuchardt <xypron.glpk@gmx.de>
 */
public class RayleighDistribution implements Distribution {

    /**
     * Random number generator
     */
    private MersenneTwister generator = new MersenneTwister();
    /**
     * Scale parameter.
     */
    private double m;
    /**
     * Mean.
     */
    private double mean;
    /**
     * Standard deviation.
     */
    private double standardDeviation;

    /**
     * Constructor.
     * @param my mean
     */
    public RayleighDistribution(final double my) {
        setMean(my);
    }

    /**
     * Set mean.
     * @param my mean
     */
    public final void setMean(final double my) {
        m = my * my * 2 / Math.PI;
        mean = my;
        standardDeviation = Math.sqrt(.5 * (4 - Math.PI) * m);
    }

    @Override
    public final double alphaSafetyFactor(final double serviceLevel)
            throws Exception {
        double increment = 16;
        double ret = 0;

        do {
            if (alphaServiceLevel(ret) > serviceLevel) {
                ret -= increment;
            } else {
                ret += increment;
            }
            increment = .5 * increment;
        } while (increment > 1e-7);

        return ret;
    }

    @Override
    public final double alphaServiceLevel(final double safetyFactor)
            throws Exception {
        double x;
        x = mean + safetyFactor * standardDeviation;
        if (x < 0) {
            return 0;
        }
        return 1. - Math.exp(-.5 * (x * x) / m);
    }

    @Override
    public final double betaSafetyFactor(double serviceLevel) throws Exception {
        double increment = 16;
        double ret = 0;

        do {
            if (betaServiceLevel(ret) > serviceLevel) {
                ret -= increment;
            } else {
                ret += increment;
            }
            increment = .5 * increment;
        } while (increment > 1e-7);

        return ret;
    }

    @Override
    public final double betaServiceLevel(final double safetyFactor)
            throws Exception {
        double x;
        double ret;
        x = mean + safetyFactor * standardDeviation;
        if (x <= 0) {
            return 0;
        }

        return Math.sqrt(0.5 * Math.PI * m) / mean
                * SpecialFunctions.erf(x / Math.sqrt(2 * m));
    }

    @Override
    public final double getMean() {
        return mean;
    }

    @Override
    public final double getStandardDeviation() {
        return standardDeviation;
    }

    @Override
    public final double randomValue() throws Exception {
        return Math.sqrt(-2. * m * Math.log(1. - generator.nextDouble()));
    }
}
