/*
 * Copyright 2012 Heinrich Schuchardt <xypron.glpk@gmx.de>.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.xypron.statistics;

/**
 *
 * @author Heinrich Schuchardt <xypron.glpk@gmx.de>
 */
public class SpecialFunctions {

    /**
     * Constructor.
     */
    private SpecialFunctions() {
    }

    /**
     * Error function.
     * Milton Abramowitz, Irene A. Stegun
     * Handbook of Mathematical Functions, 1964
     * Formula 7.1.26
     * eps < 1.5e-7
     * @param x value
     * @return error function
     */
    public static double erf(final double x) {
        final double a1 = 0.254829592;
        final double a2 = -0.284496736;
        final double a3 = 1.421413741;
        final double a4 = -1.453152027;
        final double a5 = 1.061405429;
        final double p = 0.3275911;

        double t;
        double y;
        double z;

        z = Math.abs(x);

        // A&S formula 7.1.26
        t = 1.0 / (1.0 + p * z);
        y = 1.0 - (((((a5 * t + a4) * t) + a3) * t + a2) * t + a1) * t
                * Math.exp(-z * z);
        return Math.signum(x) * y; // erf(-x) = -erf(x)
    } 

    /** Inverse error function
     * The Chebyshev coefficients were published in file erfinv.sci of Scilab.
     * http://www.scilab.org/
     *
     * @param x value
     * @return inverse error function
     */
    public static double erfInv(final double x) {

        final double[] a = {
            0.88622692374517353d,
            -1.6601283962374516d,
            0.92661860147244357d,
            -0.14110320437680104d};
        final double[] b = {
            -2.13505380615258078d,
            1.46060340345661088d,
            -0.33198239813321595d,
            0.01197270616590528d};
        final double[] c = {
            -1.994216456587148d,
            -1.87267416351196d,
            3.60874665878559364d,
            1.82365845766309853d};
        final double[] d = {
            3.74146294065960872d,
            1.81848952562894617d};
        final double pHigh = .7d;
        double sign;
        double y;
        double z;
        double r;

        y = Math.abs(x);
        sign = Math.signum(x);

        if (y > pHigh) {
            z = Math.sqrt(-Math.log((1. - y) * .5d));
            r = (((c[3] * z + c[2]) * z + c[1]) * z + c[0])
                    / ((d[1] * z + d[0]) * z + 1);
        } else {
            z = y * y;
            r = y * (((a[3] * z + a[2]) * z + a[1]) * z + a[0])
                    / ((((b[3] * z + b[2]) * z + b[1]) * z + b[0]) * z + 1);
        }

        return sign * r;
    }

}
